#include <stdio.h>    /* printf()  */
#include <stdlib.h>   /* exit(), EXIT_SUCCESS, EXIT_FAILURE */
#include <sys/io.h>   /* iopl(), outl() */

int main(void) {
	
	unsigned long pmbase_value, smi_en_ioport, smi_sts_ioport, smi_en_value, smi_sts_value;


	/* grant access to io */
	iopl(3);

	/*
	 * PMBASE, SMI_EN, SMI_STS are 32 bit register.
	 *
	 * PCI address of PMBASE: bus 0, dev 0x1F, func 0, offset 0x40
	 * in bits: 0000000011111000 + 1000000 (offset) (0xF8 + 0x40)
	 * hex: 0x138
	 *
	 * SMI_EN:  PMBASE + 0x30
	 * SMI_STS: PMBASE + 0x34
	 * 
	 * PCI configuration mechanism:
	 * 1. sending address to port 0xCF8 specifies to which device to read/write
	 * 2. reading/writing to port 0xCFC writes/reads to/from specified device in 1.
	 */

	/* read/write from PMBASE */
	outl(0xCF8, 0x138); 
	/* read value from PMBASE */
	pmbase_value = inl(0xCFC);
	
	/* PMBASE on ICH3M (my laptop) is 0x900007, in QEMU 0x0 */
	printf("PMBASE = 0x%lx\n", pmbase_value);

	smi_en_ioport =  pmbase_value + 0x30;
	smi_sts_ioport = pmbase_value + 0x34;
	printf("Address of SMI_EN  (PMBASE + 0x30): 0x%lx\n", smi_en_ioport);
	printf("Address of SMI_STS (PMBASE + 0x34): 0x%lx\n", smi_sts_ioport);

	/* read/write from SMI_EN */
	outl(0xCF8, smi_en_ioport);
	/* read value from SMI_EN */
	smi_en_value = inl(0xCFC);
	printf("Saved value in SMI_EN : 0x%lx\n", smi_en_value);

	/* read/write from SMI_STS */
	outl(0xCF8, smi_sts_ioport);
	/* read value from SMI_EN */
	smi_sts_value = inl(0xCFC);
	printf("Saved value in SMI_STS: 0x%lx\n", smi_sts_value);



	exit(EXIT_SUCCESS);


}



